#include "common.h"
#include <string.h>

const char *Source  = 
  "__kernel void Multiply(__global const uint *input,\n"
  "  __global uint *output, const uint multiplier) {\n"
  "  uint tid = get_global_id(0);\n"
  "  output[tid] = input[tid] * multiplier;\n"
  "}";

const cl_uint Multiplier = 2;
const cl_uint Width = 128;
const size_t SizeInBytes = Width * sizeof(cl_uint);
cl_uint *Input = NULL, *Output = NULL;
cl_mem InputBuffer, OutputBuffer;
cl_command_queue CommandQueue;
cl_context Context;
cl_device_id *Devices;
cl_kernel Kernel;
cl_program Program;

void InitializeHost();
void InitializeCL();
void RunCLKernels();
void CleanupCL();
void CleanupHost();
void PrintArray(const char *arrayName, 
  const cl_uint *arrayData, const cl_uint length);

int main() {
  InitializeHost();
  InitializeCL();
  RunCLKernels();
  PrintArray("Output", Output, Width);
  CleanupCL();
  CleanupHost();
  return 0;
}

void InitializeHost() {
  Input = (cl_uint*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Input);
  Output = (cl_uint*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Output);

  for(cl_uint i = 0; i < Width; i++) Input[i] = i;
  PrintArray("Input", Input, Width);
}

void PrintArray(const char *arrayName, 
  const cl_uint *arrayData, const cl_uint length) {
  
  cout << arrayName << ":" << endl;
  for(cl_uint i = 0; i < length; i++) 
    cout << arrayData[i] << " ";
  cout << endl << endl;
}

void InitializeCL() {
  cl_int err;
  cl_uint numPlatforms;
  cl_platform_id platform = NULL;
  char buff[1000];

  err = clGetPlatformIDs(0, NULL, &numPlatforms);
  CHECK_ERROR(err, "clGetPlatformIDs");
  
  if(numPlatforms > 0) {
    cl_platform_id* platforms = 
      new cl_platform_id[numPlatforms];
    err = clGetPlatformIDs(numPlatforms, platforms, NULL);
    CHECK_ERROR(err, "clGetPlatformIDs");
    for(cl_uint i = 0; i < numPlatforms; i++) {
      err = clGetPlatformInfo(platforms[i], 
        CL_PLATFORM_VENDOR, sizeof(buff), buff, NULL);
      CHECK_ERROR(err, "clGetPlatformInfo");
      if(0 == strcmp(buff, "Advanced Micro Devices, Inc.")) {
        platform = platforms[i];
        break;
      }
    }
    delete platforms;
  }
  CHECK_NOT_NULL(platform);

  cl_context_properties cps[3] = {CL_CONTEXT_PLATFORM, 
    (cl_context_properties)platform, 0};
  Context = clCreateContextFromType(cps, CL_DEVICE_TYPE_CPU,
    NULL, NULL, &err);
  CHECK_ERROR(err, "clCreateContextFromType");

  size_t deviceListSize;
  err = clGetContextInfo(Context, CL_CONTEXT_DEVICES, 0, 
    NULL, &deviceListSize);
  CHECK_ERROR(err, "clGetContextInfo");

  Devices = (cl_device_id*)malloc(deviceListSize);
  if(Devices == 0) {
    cout << __FILE__ << "(" << __LINE__ << ")" << 
      "Error : No Devices found" << endl;
    exit(EXIT_FAILURE);
  }

  err = clGetContextInfo(Context, CL_CONTEXT_DEVICES, 
    deviceListSize, Devices, NULL);
  CHECK_ERROR(err, "clGetContextInfo");

  CommandQueue = clCreateCommandQueue(Context, Devices[0], 
    0, &err);
  CHECK_ERROR(err, "clCreateCommandQueue");

  InputBuffer = clCreateBuffer(Context, 
    CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, SizeInBytes, 
    Input, &err);
  CHECK_ERROR(err, "clCreateBuffer");

  OutputBuffer = clCreateBuffer(Context, 
    CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, SizeInBytes, 
    Output, &err);
  CHECK_ERROR(err, "clCreateBuffer");

  size_t sourceSize[]  = {strlen(Source)};
  Program = clCreateProgramWithSource(Context, 1, 
    &Source, sourceSize, &err);
  CHECK_ERROR(err, "clCreateProgramWithSource");

  err = clBuildProgram(Program, 1, Devices, NULL, 
    NULL, NULL);
  CHECK_ERROR(err, "clBuildProgram");

  Kernel = clCreateKernel(Program, "Multiply", &err);
  CHECK_ERROR(err, "clCreateKernel");
}

void RunCLKernels() {
  cl_int err;
  cl_event event;
  size_t globalWorkSize[1] = {Width};
  size_t localWorkSize[1] = {Width};
  
  err = clSetKernelArg(Kernel, 0, sizeof(cl_mem), 
    (void*)&InputBuffer);
  CHECK_ERROR(err, "clSetKernelArg");
  err = clSetKernelArg(Kernel, 1, sizeof(cl_mem), 
    (void*)&OutputBuffer);
  CHECK_ERROR(err, "clSetKernelArg");
  err = clSetKernelArg(Kernel, 2, sizeof(cl_uint), 
    (void*)&Multiplier);
  CHECK_ERROR(err, "clSetKernelArg");

  err = clEnqueueNDRangeKernel(CommandQueue, Kernel, 1, 
    NULL, globalWorkSize, localWorkSize, 0, NULL, &event);
  CHECK_ERROR(err, "clEnqueueNDRangeKernel");

  err = clWaitForEvents(1, &event);
  CHECK_ERROR(err, "clWaitForEvents");
  err = clReleaseEvent(event);
  CHECK_ERROR(err, "clReleaseEvent");

  err = clEnqueueReadBuffer(CommandQueue, OutputBuffer, 
    CL_TRUE, 0, SizeInBytes, Output, 0, NULL, &event);
  CHECK_ERROR(err, "clEnqueueReadBuffer");
  
  err = clWaitForEvents(1, &event);
  CHECK_ERROR(err, "clWaitForEvents");
  err = clReleaseEvent(event);
  CHECK_ERROR(err, "clReleaseEvent");
}

void CleanupCL() {
  cl_int err;
  err = clReleaseKernel(Kernel);
  CHECK_ERROR(err, "clReleaseKernel");
  err = clReleaseProgram(Program);
  CHECK_ERROR(err, "clReleaseProgram");
  err = clReleaseMemObject(InputBuffer);
  CHECK_ERROR(err, "clReleaseMemObject");
  err = clReleaseMemObject(OutputBuffer);
  CHECK_ERROR(err, "clReleaseMemObject");
  err = clReleaseCommandQueue(CommandQueue);
  CHECK_ERROR(err, "clReleaseCommandQueue");
  err = clReleaseContext(Context);
  CHECK_ERROR(err, "clReleaseContext");
}

void CleanupHost() {
  free(Input);
  free(Output);
  free(Devices);
}
