#include "common.hpp"

const char *Source  = 
  "__kernel void Multiply(__global const uint *input,\n"
  "  __global uint *output, const uint multiplier) {\n"
  "  uint tid = get_global_id(0);\n"
  "  output[tid] = input[tid] * multiplier;\n"
  "}";
const cl_uint Multiplier = 2;
const cl_uint Width = 128;
const size_t SizeInBytes = Width * sizeof(cl_uint);
cl_uint *Input = NULL, *Output = NULL;

void PrintArray(const char *arrayName, 
  const cl_uint *arrayData, const cl_uint length);

int main() {
  cl_int err;
  cl_uint counter;
  
  Input = (cl_uint*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Input);
  Output = (cl_uint*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Output);
  for(cl_uint i = 0; i < Width; i++) Input[i] = i;
  PrintArray("Input", Input, Width);

  vector<cl::Platform> platforms;
  err = cl::Platform::get(&platforms);
  CHECK_ERROR(err, "Platform::get");

  vector<cl::Platform>::iterator i;
  if(platforms.size() > 0)
    for(i = platforms.begin(); i != platforms.end(); i++)
      if(!strcmp(
        (*i).getInfo<CL_PLATFORM_VENDOR>(&err).c_str(), 
        "Advanced Micro Devices, Inc.")) break;
  CHECK_ERROR(err, "Platform::getInfo");

  cl_context_properties cps[3] = {CL_CONTEXT_PLATFORM, 
    (cl_context_properties)(*i)(), 0};
  cl::Context context(CL_DEVICE_TYPE_CPU, cps, NULL, 
    NULL, &err);
  CHECK_ERROR(err, "Context::Context");

  vector<cl::Device> devices = 
    context.getInfo<CL_CONTEXT_DEVICES>(&err);
  CHECK_ERROR(err, "Context::getInfo");
  if (devices.size() == 0) {
    cout << "No device available" << endl;
    return EXIT_FAILURE;
  }

  size_t sourceSize  = strlen(Source);
  cl::Program::Sources sources(1, 
    make_pair(Source, sourceSize));
  cl::Program program = 
    cl::Program(context, sources, &err);
  CHECK_ERROR(err, "Program::Program");

  err = program.build(devices);
  if (err != CL_SUCCESS) {
    if(err == CL_BUILD_PROGRAM_FAILURE) {
      cl::string str = program.getBuildInfo
        <CL_PROGRAM_BUILD_LOG>(devices[0]);
      cout << "********************\n";
      cout << str.c_str() << endl;
      cout << "********************\n";
    }
  }
  CHECK_ERROR(err, "Program::build");

  cl::Kernel kernel(program, "Multiply", &err);
  CHECK_ERROR(err, "Kernel::Kernel");

  cl::Buffer inputBuffer = cl::Buffer(context, 
    CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, 
    SizeInBytes, Input, &err);
  CHECK_ERROR(err, "Buffer::Buffer");
  cl::Buffer outputBuffer = cl::Buffer(context, 
    CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, 
    SizeInBytes, Output, &err);
  CHECK_ERROR(err, "Buffer::Buffer");
  
  err = kernel.setArg(0, sizeof(cl_mem), &inputBuffer);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(1, sizeof(cl_mem), &outputBuffer);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(2, Multiplier);
  CHECK_ERROR(err, "Kernel::setArg");

  cl::CommandQueue queue(context, devices[0], 0, &err);
  CHECK_ERROR(err, "CommandQueue::CommandQueue");

  err = queue.enqueueNDRangeKernel(kernel, cl::NullRange,
    cl::NDRange(Width), cl::NDRange(Width));
  CHECK_ERROR(err, "CommandQueue::enqueueNDRangeKernel");

  err = queue.finish();
  CHECK_ERROR(err, "CommandQueue::finish");

  err = queue.enqueueReadBuffer(outputBuffer, CL_TRUE,
    0, SizeInBytes, Output);
  CHECK_ERROR(err, "CommandQueue::enqueueReadBuffer");
  
  PrintArray("Output", Output, Width);
  return 0;
}

void PrintArray(const char *arrayName, 
  const cl_uint *arrayData, const cl_uint length) {
  
  cout << arrayName << ":" << endl;
  for(cl_uint i = 0; i < length; i++) 
    cout << arrayData[i] << " ";
  cout << endl << endl;
}
