#include "common.hpp"
#include <Windows.h>

const char *Source = 
  "__kernel void MakeSum(__global const uint *input, \n"
  "  __global ulong *output, const uint width) {\n"
  "  uint tid = get_global_id(0);\n"
  "  ulong sum = 0;\n"
  "  if((tid > 0) && (tid < width)) {\n"
  "    uint min = input[tid - 1];\n"
  "    uint max = input[tid];\n"
  "    uint forAdd = min;\n"
  "    while(forAdd < max) sum += forAdd++;\n"
  "  }\n"
  "  output[tid] = sum;\n"
  "}";
const cl_uint Min = 0;
const cl_uint Width = 256;
const size_t SizeInBytes = Width * sizeof(cl_uint);
const size_t SizeInBytes64 = Width * sizeof(cl_ulong);
cl_uint *Input = NULL;
cl_ulong *Output = NULL;

void PrintArray(const char *arrayName, 
  const cl_uint *arrayData, const cl_uint length);
void PrintArray64(const char *arrayName, 
  const cl_ulong *arrayData, const cl_uint length);

int main(int argc, char *argv[]) {
  cl_ulong elapsedTime;
  cl_ulong startTime = GetTickCount();
  cl_int err = 0;

  if(argc != 2) {
    cout << "1 arg is needed\n";
    return 1;
  }
  cl_uint max = _atoi64(argv[1]);

  Input = (cl_uint*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Input);
  Output = (cl_ulong*)malloc(SizeInBytes64);
  CHECK_NOT_NULL(Output);

  cl_uint range = (max - Min) / (Width - 1);
  cl_uint midValue = Min + range;
  Input[0] = Min;
  Input[Width - 1] = max;
  for(cl_uint i = 1; i < Width - 1; i++) {
    Input[i] = midValue;
    midValue += range;
  }
  //PrintArray("Input", Input, Width);
  
  vector<cl::Platform> platforms;
  err = cl::Platform::get(&platforms);
  CHECK_ERROR(err, "Platform::get");

  vector<cl::Platform>::iterator i;
  if(platforms.size() > 0)
    for(i = platforms.begin(); i != platforms.end(); i++)
      if(!strcmp(
        (*i).getInfo<CL_PLATFORM_VENDOR>(&err).c_str(),
        "Advanced Micro Devices, Inc.")) break;
  CHECK_ERROR(err, "Platform::getInfo");
  cl_context_properties cps[3] = {CL_CONTEXT_PLATFORM, 
    (cl_context_properties)(*i)(), 0};

  cl::Context context(CL_DEVICE_TYPE_CPU, cps, 
    NULL, NULL, &err);
  CHECK_ERROR(err, "Context::Context");

  vector<cl::Device> devices = 
    context.getInfo<CL_CONTEXT_DEVICES>();
  CHECK_ERROR(err, "Context::getInfo");
  if (devices.size() == 0) {
    cout << "No device available" << endl;
    return EXIT_FAILURE;
  }

  size_t sourceSize  = strlen(Source);
  cl::Program::Sources sources(1, 
    make_pair(Source, sourceSize));
  cl::Program program = 
    cl::Program(context, sources, &err);
  CHECK_ERROR(err, "Program::Program");

  err = program.build(devices);
  if (err != CL_SUCCESS) {
    if(err == CL_BUILD_PROGRAM_FAILURE) {
      cl::string str = program.getBuildInfo
        <CL_PROGRAM_BUILD_LOG>(devices[0]);
      cout << "********************\n";
      cout << str.c_str() << endl;
      cout << "********************\n";
    }
  }
  CHECK_ERROR(err, "Program::build");

  cl::Kernel kernel(program, "MakeSum", &err);
  CHECK_ERROR(err, "Kernel::Kernel");

  cl::Buffer inputBuffer = cl::Buffer(context, 
    CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, 
    SizeInBytes, Input, &err);
  CHECK_ERROR(err, "Buffer::Buffer");
  cl::Buffer outputBuffer = cl::Buffer(context, 
    CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, 
    SizeInBytes64, Output, &err);
  CHECK_ERROR(err, "Buffer::Buffer");

  err = kernel.setArg(0, sizeof(cl_mem), &inputBuffer);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(1, sizeof(cl_mem), &outputBuffer);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(2, Width);
  CHECK_ERROR(err, "Kernel::setArg");

  cl::CommandQueue queue(context, devices[0], 0, &err);
  CHECK_ERROR(err, "CommandQueue::CommandQueue");

  err = queue.enqueueNDRangeKernel(kernel, cl::NullRange,
    cl::NDRange(Width), cl::NDRange(Width));
  CHECK_ERROR(err, "CommandQueue::enqueueNDRangeKernel");

  err = queue.finish();
  CHECK_ERROR(err, "CommandQueue::finish");

  err = queue.enqueueReadBuffer(outputBuffer, CL_TRUE, 
    0, SizeInBytes64, Output);
  CHECK_ERROR(err, "CommandQueue::enqueueReadBuffer");
  //PrintArray64("Output", Output, Width);
  
  cl_ulong sum = 0;
  for(cl_uint i = 1; i < Width; i++) sum += Output[i];
  sum += max;
  
  free(Input);
  free(Output);
  elapsedTime = GetTickCount() - startTime;
  cout << Min << '\t' << max << '\t' << sum << '\t' 
    << elapsedTime << endl;
  return 0;
}

void PrintArray(const char *arrayName, 
  const cl_uint *arrayData, const cl_uint length) {
  cout << arrayName << ":" << endl;
  for(cl_uint i = 0; i < length; i++) 
    cout << arrayData[i] << " ";
  cout << endl << endl;
}

void PrintArray64(const char *arrayName, 
  const cl_ulong *arrayData, const cl_uint length) {
  cout << arrayName << ":" << endl;
  for(cl_uint i = 0; i < length; i++) 
    cout << arrayData[i] << " ";
  cout << endl << endl;
}
