#include "common.hpp"
#include <Windows.h>
#include <math.h>

const char *Source  = 
  "#pragma OPENCL EXTENSION cl_amd_fp64 : enable\n"
  //"#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n"
  "__kernel void GetArea(__global const double *input,\n"
  "  __global double *output, const uint width,\n"
  "  const uint divide) {\n"
  "\n"
  "  uint tid = get_global_id(0);\n"
  "  if((tid > 0) && (tid < width)) {\n"
  "    const double min = input[tid - 1];\n"
  "    const double max = input[tid];\n"
  "    const double dividedX = (max - min) / \n"
  "      (double)divide;\n"
  "    double x = min + dividedX / 2;\n"
  "    double y;\n"
  "    double area = 0.0;\n"
  "    for(uint i = 0; i < divide; i++) {\n"
  "      y = -0.006 * (x * x * x) + \n"
  "        0.1 * (x * x) + 20.0 * x + 1000.0;\n"
  "      area += fabs(y) * dividedX;\n"
  "      x += dividedX;\n"
  "    }\n"
  "    output[tid] = area;\n"
  "  }\n"
  "}";

void PrintArray(const char *arrayName, 
  const double *arrayData, const cl_uint length);

const double Min = 0.0;
const double Max= 50.0;
const cl_uint Width = 250;
const size_t SizeInBytes = Width * sizeof(double);
double *Input = NULL;
double *Output = NULL;

int main(int argc, char *argv[]) {
  cl_int err = 0;
  cl_ulong elapsedTime;
  cl_ulong startTime = GetTickCount();

  if(argc != 2) {
    cout << "1 arg is needed\n";
    return 1;
  }
  cl_uint divide = atoi(argv[1]);
  
  Input = (double*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Input);
  Output = (double*)malloc(SizeInBytes);
  CHECK_NOT_NULL(Output);

  const double range = (Max - Min) / (double)(Width - 1);
  double midValue = Min + range;
  Input[0] = Min;
  Input[Width - 1] = Max;
  
  for(cl_uint i = 1; i < Width - 1; i++) {
    Input[i] = midValue;
    midValue += range;
  }
  //PrintArray("Input", Input, Width);
  
  vector<cl::Platform> platforms;
  err = cl::Platform::get(&platforms);
  CHECK_ERROR(err, "Platform::get");

  vector<cl::Platform>::iterator i;
  if(platforms.size() > 0)
    for(i = platforms.begin(); i != platforms.end(); i++)
      if(!strcmp(
        (*i).getInfo<CL_PLATFORM_VENDOR>(&err).c_str(), 
        "Advanced Micro Devices, Inc.")) break;
  CHECK_ERROR(err, "Platform::getInfo");
  cl_context_properties cps[3] = {CL_CONTEXT_PLATFORM, 
    (cl_context_properties)(*i)(), 0};

  cl::Context context(CL_DEVICE_TYPE_CPU, cps, NULL, 
    NULL, &err);
  CHECK_ERROR(err, "Context::Context");

  vector<cl::Device> devices = 
    context.getInfo<CL_CONTEXT_DEVICES>();
  CHECK_ERROR(err, "Context::getInfo");
  if (devices.size() == 0) {
    cout << "No device available" << endl;
    return EXIT_FAILURE;
  }

  size_t sourceSize  = strlen(Source);
  cl::Program::Sources sources(1, 
    make_pair(Source, sourceSize));
  cl::Program program = 
    cl::Program(context, sources, &err);
  CHECK_ERROR(err, "Program::Program");

  err = program.build(devices);
  if (err != CL_SUCCESS) {
    if(err == CL_BUILD_PROGRAM_FAILURE) {
      cl::string str = 
        program.getBuildInfo<CL_PROGRAM_BUILD_LOG>
          (devices[0]);
      cout << "********************\n";
      cout << str.c_str() << endl;
      cout << "********************\n";
    }
  }
  CHECK_ERROR(err, "Program::build");

  cl::Kernel kernel(program, "GetArea", &err);
  CHECK_ERROR(err, "Kernel::Kernel");

  cl::Buffer inputBuffer = cl::Buffer(context, 
    CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, 
    SizeInBytes, Input, &err);
  CHECK_ERROR(err, "Buffer::Buffer");
  cl::Buffer outputBuffer = cl::Buffer(context, 
    CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, 
    SizeInBytes, Output, &err);
  CHECK_ERROR(err, "Buffer::Buffer");

  err = kernel.setArg(0, sizeof(cl_mem), 
    &inputBuffer);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(1, sizeof(cl_mem), 
    &outputBuffer);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(2, Width);
  CHECK_ERROR(err, "Kernel::setArg");
  err = kernel.setArg(3, divide);
  CHECK_ERROR(err, "Kernel::setArg");

  cl::CommandQueue queue(context, devices[0], 0, &err);
  CHECK_ERROR(err, "CommandQueue::CommandQueue");

  err = queue.enqueueNDRangeKernel(kernel, cl::NullRange,
    cl::NDRange(Width), cl::NDRange(Width));
  CHECK_ERROR(err, "CommandQueue::enqueueNDRangeKernel");

  err = queue.finish();
  CHECK_ERROR(err, "CommandQueue::finish");

  err = queue.enqueueReadBuffer(outputBuffer, CL_TRUE, 0,
    SizeInBytes, Output);
  CHECK_ERROR(err, "CommandQueue::enqueueReadBuffer");
  //PrintArray("Output", Output, Width);
  
  double area = 0.0;
  for(cl_uint i = 1; i < Width; i++) area += Output[i];
  
  free(Input);
  free(Output);
  elapsedTime = GetTickCount() - startTime;
  cout.setf(ios::fixed);
  cout.width(14);
  cout << (cl_ulong)divide * Width << '\t' << area 
    << '\t' << elapsedTime << endl;
  return 0;
}

void PrintArray(const char *arrayName, 
  const double *arrayData, const cl_uint length) {
  
  cout << arrayName << ":" << endl;
  for(cl_uint i = 0; i < length; i++) 
    cout << arrayData[i] << " ";
  cout << endl << endl;
}
